import logging

from abc import ABC, abstractmethod
from typing import Any, Optional, List, Set

from inds.core.consumer import Consumer
from inds.core.item import Item
from inds.core.plugin import Plugin
from inds.core.producer import Producer, Property

logger = logging.getLogger(__name__)


class SimplePlugin(Plugin, ABC):
    def __init__(self, name: str, properties: Optional[List[Property]] = None,
                 input_queue_size=1) -> None:
        super(SimplePlugin, self).__init__(name)

        self.producer: Producer = Producer(name, properties)
        self.add_producer(self.producer)

        self.consumer: Consumer = Consumer(name, input_queue_size)
        self.add_consumer(self.consumer)

    @abstractmethod
    def process_item(self, item: Item) -> Item:
        """Process a single item and return it"""

    def on_output_dropped(self, item: Item, consumers: Set[str]) -> None:
        """
        Evoked whenever output is dropped at at least one consumer.

        :param item:       The item that was dropped
        :param consumers:  The consumers that did not take the item.
        """
        pass

    def on_stopped(self, exception: Optional[BaseException] = None) -> Optional[bool]:
        """
        Callback that is evoked when run terminates. Can be implemented by a derived class.

        :param exception:  Exception that was raised and terminates this the run function or None if the run
                           exited gracefully.

        :return:           True if the exception is not to to be raised.
                           Otherwise (i.e., if the return value is None, False or of any type except bool),
                           the exception is thrown immediately after on_stopped exits.
        """
        return False

    def forward_item(self, item: Item, **kwargs):
        # Forward the processed result if any.
        if item is not None:
            dropped = self.producer.send(item, **kwargs)

            if len(dropped):
                self.on_output_dropped(item, dropped)

    def run(self) -> None:
        ex: Optional[BaseException] = None

        try:
            while self.running:
                item: Item = self.consumer.receive()
                if item is not None:
                    self.forward_item(self.process_item(item))

        except BaseException as e:
            ex = e

        finally:
            noexcept_exit: Any = self.on_stopped(ex)

            if ex is not None and noexcept_exit is not None and type(noexcept_exit) != bool:
                logger.warning(f'Return value of `on_stopped` implementation is not bool: {noexcept_exit}. '
                               f'Raising exception independent of true value of {noexcept_exit}.')

            if ex is not None and (type(noexcept_exit) != bool or not noexcept_exit or not isinstance(e, Exception)):
                raise ex
