from abc import ABCMeta
from typing import Tuple, Optional, List, Dict

import cv2
import numpy as np

from inds.inference.models.base_model import BaseModel, InferenceMode


class ImageBaseModel(BaseModel, metaclass=ABCMeta):

    def __init__(self, input_shape: Tuple[int, int, int, int] = (0, 0, 0, 0),
                 inference_mode: InferenceMode = InferenceMode.WHOLE_IMAGE,
                 selected_rois: Optional[List[str]] = None):
        self.input_shape: Tuple[int, int, int, int] = input_shape
        self.inference_mode: InferenceMode = inference_mode
        self.selected_rois: Optional[List[str]] = selected_rois
        self.current_roi_order: List[str] = []

    def preprocess(self, item: Dict):
        image = item.get('image')
        if image is None:
            return None

        height, width, channels = image.shape

        image_patches = []
        if self.inference_mode == InferenceMode.WHOLE_IMAGE:
            image_patches.append(image)
        elif self.inference_mode == InferenceMode.ALL_ROIS or self.inference_mode == InferenceMode.SELECTED_ROIS:
            item_rois = item.get('rois', {})
            selected_rois = None
            self.current_roi_order = []
            if self.inference_mode == InferenceMode.ALL_ROIS:
                selected_rois = item_rois.keys()
            elif self.inference_mode == InferenceMode.SELECTED_ROIS:
                selected_rois = self.selected_rois

            if selected_rois is not None:
                for roi_name in selected_rois:
                    roi = item_rois.get(roi_name)
                    if roi is None:
                        continue
                    x_left, x_right, y_top, y_bottom = roi
                    if 0 <= x_left < x_right <= width and 0 <= y_top < y_bottom <= height:
                        image_patch = image[y_top:y_bottom, x_left:x_right, :]
                        image_patches.append(image_patch)
                        self.current_roi_order.append(roi_name)

        if len(image_patches) == 0:
            return None

        images = np.zeros((len(image_patches), self.input_shape[1], self.input_shape[2], self.input_shape[3]),
                          dtype=np.float32)
        for i, image_patch in enumerate(image_patches):
            # resize
            img = cv2.resize(image_patch, (self.input_shape[3], self.input_shape[2]))

            # transpose from HWC to CHW
            img = img.transpose((2, 0, 1))

            # convert to float
            # image = np.array(image, dtype=np.float32)
            img = img.astype(np.float32)

            # normalize: image = (2.0 / 255.0) * image - 1
            img *= 2.0 / 255.0
            img -= 1

            images[i] = img

        return images
