"""
THIS FILE ORIGINATES FROM SICK AG.

This module handles the incoming data and extracts information from it.

Copyright note: Redistribution and use in source, with or without modification, are permitted.

created: 2015-01-19
updated: 2018-02-02

author: Uwe Hahne, Torben Peichl
SICK AG, Waldkirch
email: techsupport0905@sick.de
Last commit: $Date: 2018-02-02 12:31:38 +0100 (Fr, 02 Feb 2018) $
Last editor: $Author: peichto $

Version: "$Revision: 15773 $"

"""

import logging
import xml.etree.ElementTree as ET
from struct import unpack, calcsize
from typing import Optional

import numpy as np

logger = logging.getLogger(__name__)

# ----------------------------------------------------------------------------------------------------------------------
MAX_CONFIDENCE = 65535


# ----------------------------------------------------------------------------------------------------------------------

class Data:
    """ Gathers methods to handle the raw data. """

    def __init__(self, xml_parser=None, changed_counter=-1, depthmap=None, polar_data=None, checksum='E'):
        self.xml_parser = xml_parser
        self.changed_counter = changed_counter
        self.checksum = checksum
        self.corrupted = False

        self.parameters_changed = True
        self.camera_parameters: Optional[CameraParameters] = None

        self.depthmap = depthmap
        self.polar_data_2d = polar_data
        self.cartesian_data = None

        self.has_depth_map = False
        self.has_polar_2d = False
        self.has_cartesian = False

    def read(self, data_buffer):
        """ Extracts necessary data segments and triggers parsing of segments. """

        # first 11 bytes contain some internal definitions
        temp_buffer = data_buffer[0:11]
        (magicword, pkglength, protocol_version, packet_type) = \
            unpack('>IIHB', temp_buffer)
        assert (magicword == 0x02020202)
        logger.debug("Package length: %s", pkglength)
        logger.debug("Protocol version: %s", protocol_version)  # expected to be == 1
        logger.debug("Packet type: %s", packet_type)  # expected to be  == 98

        # next four bytes an id (should equal 1) and
        # the number of segments (should be 3)
        temp_buffer = data_buffer[11:15]
        (segid, num_segments) = unpack('>HH', temp_buffer)
        logger.debug("Blob ID: %s", segid)  # expected to be == 1
        logger.debug("Number of segments: %s", num_segments)  # expected to be  == 3

        # offset and changedCounter, 4 bytes each per segment
        offset = [None] * num_segments
        changed_counter = [None] * num_segments
        temp_buffer = data_buffer[15:15 + num_segments * 2 * 4]

        for i in range(num_segments):
            index = i * 8
            (offset[i], changed_counter[i]) = \
                unpack('>II', temp_buffer[index:index + 8])
            offset[i] += 11

        logger.debug("Offsets: %s", offset)  # offset in bytes for each segment
        logger.debug("Changed counter: %s", changed_counter)  # counter for changes in the data

        # first segment describes the data format in XML
        xml_segment = data_buffer[offset[0]:offset[1]]
        logger.debug("The whole XML segment:")
        logger.debug(xml_segment)
        # second segment contains the binary data
        binary_segment = data_buffer[offset[1]:offset[2]]

        if num_segments == 3:
            # numBytes(magicword) = 4, numBytes(pkglength) = 4
            overlaySegment = data_buffer[offset[2]:pkglength + 4 + 4]
            logger.debug("The whole overlay XML segment:")
            logger.debug(overlaySegment)

        checksum = data_buffer[pkglength + 8]
        if checksum != self.checksum:
            logger.error("Checksum is wrong: %s (expected %s)" % (checksum, self.checksum))  # checksum of whole data
            self.corrupted = True
        else:
            logger.debug("Checksum: %s", checksum)  # checksum of whole data
            self.corrupted = False

        # parsing the XML in order to extract necessary image information
        # only parse if something has changed
        self.parameters_changed = self.changed_counter < changed_counter[0]
        if self.parameters_changed:
            logger.debug("XML did change, parsing started.")
            my_xml_parser = XMLParser()
            my_xml_parser.parse(xml_segment)
            self.xml_parser = my_xml_parser
            self.changed_counter = changed_counter[0]
        else:
            logger.debug("XML did not change, not parsing again.")
            my_xml_parser = self.xml_parser

        my_binary_parser = BinaryParser()

        self.has_depth_map = False
        self.has_polar_2d = False
        self.has_cartesian = False

        if my_xml_parser.hasDepthMap:
            logger.debug("Data contains depth map, reading camera params")
            self.has_depth_map = True
            self.camera_parameters = \
                CameraParameters(width=my_xml_parser.imageWidth,
                                 height=my_xml_parser.imageHeight,
                                 cam2worldMatrix=my_xml_parser.cam2worldMatrix,
                                 fx=my_xml_parser.fx, fy=my_xml_parser.fy,
                                 cx=my_xml_parser.cx, cy=my_xml_parser.cy,
                                 k1=my_xml_parser.k1, k2=my_xml_parser.k2,
                                 f2rc=my_xml_parser.f2rc)

            # extracting data from the binary segment (distance, intensity
            # and confidence).
            if my_xml_parser.stereo:
                num_bytes_distance = my_xml_parser.imageHeight * \
                                     my_xml_parser.imageWidth * \
                                     my_xml_parser.numBytesPerZValue
            else:
                num_bytes_distance = my_xml_parser.imageHeight * \
                                     my_xml_parser.imageWidth * \
                                     my_xml_parser.numBytesPerDistanceValue
            num_bytes_intensity = my_xml_parser.imageHeight * \
                                  my_xml_parser.imageWidth * \
                                  my_xml_parser.numBytesPerIntensityValue
            num_bytes_confidence = my_xml_parser.imageHeight * \
                                   my_xml_parser.imageWidth * \
                                   my_xml_parser.numBytesPerConfidenceValue

            try:
                num_bytes_frame_number = my_xml_parser.numBytesFrameNumber
                num_bytes_quality = my_xml_parser.numBytesQuality
                num_bytes_status = my_xml_parser.numBytesStatus
            except AttributeError:
                num_bytes_frame_number = 0
                num_bytes_quality = 0
                num_bytes_status = 0

            my_binary_parser.getDepthMap(binary_segment,
                                         num_bytes_frame_number,
                                         num_bytes_quality,
                                         num_bytes_status,
                                         num_bytes_distance,
                                         num_bytes_intensity,
                                         my_xml_parser.numBytesPerIntensityValue,
                                         num_bytes_confidence)

            # if myXMLParser.stereo:
            #     distance = list(myBinaryParser.depthmap.distance)
            #     for i in range(0, len(distance)):
            #         distance[i] = distance[i] / 10.0 #account for sub-millimeter values
            #     myBinaryParser.depthmap.distance = tuple(distance)
            self.depthmap = my_binary_parser.depthmap

        if my_xml_parser.hasPolar2DData:
            self.has_polar_2d = True
            if my_xml_parser.hasDepthMap:
                my_binary_parser.getPolar2D(my_binary_parser.remainingBuffer, my_xml_parser.numPolarValues)
            else:
                my_binary_parser.getPolar2D(binary_segment, my_xml_parser.numPolarValues)
            if hasattr(my_binary_parser, 'polardata'):
                self.polar_data_2d = my_binary_parser.polardata
            else:
                self.has_polar_2d = False
        elif my_xml_parser.hasCartesianData:
            self.has_cartesian = True
            if my_xml_parser.hasDepthMap:
                my_binary_parser.getCartesian(my_binary_parser.remainingBuffer)
            else:
                my_binary_parser.getCartesian(binary_segment)
            if hasattr(my_binary_parser, 'cartesianData'):
                self.cartesian_data = my_binary_parser.cartesianData
            else:
                self.has_cartesian = False


class XMLParser:
    """ The XML parser that only extracts the needed information. """

    def parse(self, xmlString):
        """ Parse method needs the XML segment as string input. """

        sickRecord = ET.fromstring(xmlString)  # the whole block set
        # is called sickrecord
        self.hasDepthMap = False
        self.hasPolar2DData = False
        self.hasCartesianData = False
        # extract camera/image parameters directly from XML
        # =======================================================================

        self.stereo = (sickRecord.find('DataSets/DataSetStereo') is not None)
        logger.info(" Blob-XML contains DataSets/DataSetStereo ? -> {}".format(self.stereo))

        for dataSetDepthMap in sickRecord.iter('DataSetStereo' if self.stereo else 'DataSetDepthMap'):
            self.availableFrames = int(dataSetDepthMap.get('datacount'))
            self.binFileName = dataSetDepthMap.find('DataLink/FileName').text
            self.hasDepthMap = True

            self.f2rc = 0.0

            for formatDescriptionDepthMap in dataSetDepthMap.iter('FormatDescriptionDepthMap'):
                for width in formatDescriptionDepthMap.iter('Width'):
                    self.imageWidth = int(width.text)
                for height in formatDescriptionDepthMap.iter('Height'):
                    self.imageHeight = int(height.text)
                for cameraToWorldTransform in formatDescriptionDepthMap.iter('CameraToWorldTransform'):
                    self.cam2worldMatrix = []  # init array
                    for idx, child in enumerate(cameraToWorldTransform):
                        self.cam2worldMatrix.append(float(child.text))  # fill array row by row (4x4 matrix)
                for cameraMatrix in formatDescriptionDepthMap.iter(
                        'CameraMatrix'):  # camera intrinsics (fx,fy,cx,cy) are the sixth child in datastream
                    self.fx = float(cameraMatrix.find('FX').text)
                    self.fy = float(cameraMatrix.find('FY').text)
                    self.cx = float(cameraMatrix.find('CX').text)
                    self.cy = float(cameraMatrix.find('CY').text)
                for distortionParams in formatDescriptionDepthMap.iter('CameraDistortionParams'):
                    self.k1 = float(distortionParams.find('K1').text)  # we only need k1 and k2
                    self.k2 = float(distortionParams.find('K2').text)
                for f2rc in formatDescriptionDepthMap.iter('FocalToRayCross'):
                    self.f2rc = float(f2rc.text)
                for frameNumber in formatDescriptionDepthMap.iter('FrameNumber'):
                    if frameNumber.text.lower() == 'uint32':
                        self.numBytesFrameNumber = 4
                for quality in formatDescriptionDepthMap.iter('Quality'):
                    if quality.text.lower() == 'uint8':
                        self.numBytesQuality = 1
                for status in formatDescriptionDepthMap.iter('Status'):
                    if status.text.lower() == 'uint8':
                        self.numBytesStatus = 1
                if self.stereo:
                    for Z in formatDescriptionDepthMap.iter('Z'):
                        self.distType = Z.text
                        if Z.text.lower() == 'uint16':
                            self.numBytesPerZValue = 2
                            self.numBytesPerDistanceValue = 2  # leagacy for ssr-loader
                else:
                    for distance in formatDescriptionDepthMap.iter('Distance'):
                        self.distType = distance.text
                        if distance.text.lower() == 'uint16':
                            self.numBytesPerDistanceValue = 2
                for intensity in formatDescriptionDepthMap.iter('Intensity'):
                    self.intsType = intensity.text
                    if intensity.text.lower() == 'uint16':
                        self.numBytesPerIntensityValue = 2
                    elif intensity.text.lower() == 'uint32':
                        self.numBytesPerIntensityValue = 4
                for confidence in formatDescriptionDepthMap.iter('Confidence'):
                    self.confType = confidence.text
                    if confidence.text.lower() == 'uint16':
                        self.numBytesPerConfidenceValue = 2
        # =======================================================================
        for dataSetPolar2D in sickRecord.iter('DataSetPolar2D'):
            self.hasPolar2DData = True
            for formatDescription in dataSetPolar2D.iter('FormatDescription'):
                for dataStream in formatDescription.iter('DataStream'):
                    if dataStream.attrib.get('type') == 'distance':
                        self.numPolarValues = int(dataStream.attrib.get('datalength'))
        # =======================================================================
        for dataSetCartesian in sickRecord.iter('DataSetCartesian'):
            self.hasCartesianData = True
            for formatDescriptionCartesian in dataSetCartesian.iter('FormatDescriptionCartesian'):
                for dataStream in formatDescriptionCartesian.iter('DataStream'):
                    for length in dataStream.iter('Length'):
                        assert (length.text.lower() == 'uint32')
                    for x in dataStream.iter('X'):
                        assert (x.text.lower() == 'float32')
                    for y in dataStream.iter('Y'):
                        assert (y.text.lower() == 'float32')
                    for z in dataStream.iter('Z'):
                        assert (z.text.lower() == 'float32')
                    for confidence in dataStream.iter(
                            'Intensity'):  # 'Intensity is not a typo, we abuse this element in the format for confidence values
                        assert (confidence.text.lower() == 'float32')


class BinaryParser:
    """ The binary parser for extracting distance, intensity and confidence from
    the binary segment of the raw data frame.
    """

    def getDepthMap(self,
                    binarySegment,
                    numBytesFrameNumber,
                    numBytesQuality,
                    numBytesStatus,
                    numBytesDistance,
                    numBytesIntensity,
                    numBytesPerIntensityValue,
                    numBytesConfidence):
        position = 0
        # the binary part starts with entries for length, a timestamp
        # and a version identifier
        infoBlockSize = calcsize('<IQH')
        (lengthAtStart, timeStamp, version) = unpack('<IQH', binarySegment[position:position + infoBlockSize])
        position += infoBlockSize
        logger.debug("Length at start: %s", lengthAtStart)
        self.logTimeStamp(timeStamp)
        logger.debug("Format version: %s", version)

        format2BlockSize = 0
        if version == 2:
            assert numBytesFrameNumber == 4
            assert numBytesQuality == 1
            assert numBytesStatus == 1
            format2BlockSize = calcsize('<IBB')
            (frameNumber, quality, status) = unpack('<IBB', binarySegment[position:position + format2BlockSize])
            position += format2BlockSize
            logger.debug("FrameNumber: %s", frameNumber)
            logger.debug("Data quality: %s", quality)
            logger.debug("Device status: %s", status)
        else:
            logger.warning("Old format, no values for frameNumber, quality and status")
            frameNumber = -1
            quality = 0
            status = 0

        dataBlockSize = numBytesDistance + \
                        numBytesIntensity + \
                        numBytesConfidence  # calculating the end index
        dataBinary = binarySegment[position:position + dataBlockSize]  # whole data block
        position += dataBlockSize
        distance = dataBinary[0:numBytesDistance]  # only the distance data (as string)

        logger.debug("Reading distance...")
        # distanceData = unpack('<%uH' % (len(distance) / 2), distance)
        distanceData = np.frombuffer(distance, dtype=np.uint16) * 0.1
        logger.debug("...done.")

        # extract the intensity data (same procedure as distance)
        logger.debug("Reading intensity...")
        off = numBytesDistance
        intensity = dataBinary[off:numBytesIntensity + off]
        # if numBytesPerIntensityValue == 2:
        #     intensityData = unpack('<%uH' % (len(intensity) / 2), intensity)
        # elif numBytesPerIntensityValue == 4:
        #     intensityData = unpack('<%uL' % (len(intensity) / 4), intensity)
        # else:
        # legacy mode, also used for RGBA -> byte-wise
        # intensityData = unpack('<%uB' % len(intensity), intensity)
        intensityData = np.frombuffer(intensity, dtype=np.uint8)
        logger.debug("...done.")

        # extract the confidence data (same procedure as distance)
        logger.debug("Reading confidence...")
        off += numBytesIntensity
        confidence = dataBinary[off:numBytesConfidence + off]
        confidenceData = unpack('<%uH' % (len(confidence) / 2), confidence)
        logger.debug("...done.")

        # checking if all data is read
        if (position + 4 == lengthAtStart):
            check = calcsize('<II')
            (crc, lengthAtEnd) = unpack('<II', binarySegment[position:position + check])
            position += check
            logger.debug("Length at start: %s", lengthAtStart)
            logger.debug("Length at end: %s", lengthAtEnd)
            # assert lengthAtStart == lengthAtEnd
            if lengthAtStart != lengthAtEnd:
                logger.error("lengthAtStart != lengthAtEnd")
        self.remainingBuffer = binarySegment[position:]

        self.depthmap = DepthMap(distanceData, intensityData, confidenceData, frameNumber, quality, status, timeStamp)

    def getPolar2D(self,
                   binarySegment,
                   numPolarValues):
        position = 0
        infoBlockSize = calcsize('<IQHIIffffff')
        if len(binarySegment) < infoBlockSize:
            logger.warning("Found inconsitency in binary polar data.")
            return
        (lengthAtStart, timeStamp, deviceID, scanCounter, syscountScan, scanFrequency, measFrequency,
         angleFirstScanPoint, angularResolution, scale,
         offset) = unpack('<IQHIIffffff', binarySegment[position:position + infoBlockSize])
        logger.debug("Length = %i" % lengthAtStart)
        position += infoBlockSize

        self.logTimeStamp(timeStamp)
        logger.debug("DeviceID = %i" % deviceID)
        logger.debug("ScanCounter = %i" % scanCounter)
        logger.debug("SyscountScan = %i" % syscountScan)
        logger.debug("ScanFrequency = %i" % scanFrequency)
        logger.debug("MeasFrequency = %i" % measFrequency)
        logger.debug("AngleFirstScanPoint = %i" % angleFirstScanPoint)
        logger.debug("AngularResolution = %i" % angularResolution)
        logger.debug("Scale = %i" % scale)
        logger.debug("Offset = %i" % offset)

        logger.debug("Reading position is now: %i" % position)
        # distanceDataSize = int(singleValueSize * numPolarValues)
        endPosition = position + numPolarValues * calcsize('<f')
        if len(binarySegment) < endPosition:
            logger.warning("Found inconsitency in binary polar data.")
            return
        distanceData = unpack('<%uf' % numPolarValues, binarySegment[position:endPosition])
        logger.debug("Distance data = %s" % str(distanceData))
        position = endPosition
        logger.debug("Reading position is now: %i" % position)

        confidenceBlockSize = calcsize('<ffff')
        endPosition = position + confidenceBlockSize
        if len(binarySegment) < endPosition:
            logger.warning("Found inconsitency in binary polar data.")
            return
        (rssi_startAngle, rssi_angularResolution, rssi_scale, rssi_offset) = unpack('<ffff',
                                                                                    binarySegment[position:endPosition])
        logger.debug("RSSI AngleFirstScanPoint = %i" % rssi_startAngle)
        logger.debug("RSSI AngularResolution = %i" % rssi_angularResolution)
        logger.debug("RSSI Scale = %i" % rssi_scale)
        logger.debug("RSSI Offset = %i" % rssi_offset)

        position = endPosition
        logger.debug("Reading position is now: %i" % position)

        endPosition = position + numPolarValues * calcsize('<f')
        if len(binarySegment) < endPosition:
            logger.warning("Found inconsitency in binary polar data.")
            return
        confidenceData = unpack('<%uf' % numPolarValues, binarySegment[position:endPosition])
        logger.debug("Confidence data = %s" % str(confidenceData))
        # convert into percent if needed
        # confidence = tuple((val / MAX_CONFIDENCE) * 100.0 for val in confidenceData)
        position = endPosition
        logger.debug("Reading position is now: %i" % position)

        # checking if all data is read
        if (position + 8 == lengthAtStart):
            check = calcsize('<II')
            endPosition = position + check
            if len(binarySegment) < endPosition:
                logger.warning("Found inconsitency in binary polar data.")
                return
            (crc, lengthAtEnd) = unpack('<II', binarySegment[position:endPosition])
            logger.debug("Length at start: %s", lengthAtStart)
            logger.debug("Length at end: %s", lengthAtEnd)
            assert lengthAtStart == lengthAtEnd
            self.hasRemainingBuffer = False
        else:
            # there is another data set in this binary buffer
            self.hasRemainingBuffer = True
            self.remainingBuffer = binarySegment[position:lengthAtStart + 4]
            self.length = lengthAtStart  # checking if all data is read

        self.polardata = Polar2DData(angleFirstScanPoint, angularResolution, distanceData, confidenceData, timeStamp)

    def getCartesian(self,
                     binarySegment):
        position = 0
        infoBlockSize = calcsize('<IQHI')
        if len(binarySegment) < infoBlockSize:
            logger.warning("Found inconsitency in binary polar data.")
            return
        (lengthAtStart, timeStamp, version, numPoints) = unpack('<IQHI',
                                                                binarySegment[position:position + infoBlockSize])
        logger.debug("Length = %i" % lengthAtStart)
        position += infoBlockSize

        self.logTimeStamp(timeStamp)
        logger.debug("Version = %i" % version)

        logger.debug("Reading position is now: %i" % position)

        endPosition = position + numPoints * calcsize('<ffff')  # 4 float32 values per point
        if len(binarySegment) < endPosition:
            logger.warning("Found inconsitency in binary polar data.")
            return
        pointCloudData = unpack('<%uf' % (numPoints * 4), binarySegment[position:endPosition])
        x = pointCloudData[0:numPoints * 4:4]
        y = pointCloudData[1:numPoints * 4:4]
        z = pointCloudData[2:numPoints * 4:4]
        rssi = pointCloudData[3:numPoints * 4:4]
        confidence = tuple((val / MAX_CONFIDENCE) * 100.0 for val in rssi)

        position = endPosition
        logger.debug("Reading position is now: %i" % position)
        logger.debug("Point cloud data = %s" % str(pointCloudData))

        # checking if all data is read
        check = calcsize('<II')
        endPosition = position + check
        if len(binarySegment) < endPosition:
            logger.warning("Found inconsitency in binary Cartesian data.")
            return
        (crc, lengthAtEnd) = unpack('<II', binarySegment[position:endPosition])
        logger.debug("Length at start: %s", lengthAtStart)
        logger.debug("Length at end: %s", lengthAtEnd)
        assert lengthAtStart == lengthAtEnd
        self.hasRemainingBuffer = False

        self.cartesianData = CartesianData(numPoints, x, y, z, confidence, timeStamp)

    # ===============================================================================

    def logTimeStamp(self, timeStamp):
        # 0x03 D9 08 40 02 C7 B0 00
        # 0000 0011 1101 1001 0000 1000 0100 0000 0000 0010 1100 0111 1011 0000 0000 0000
        # .... .YYY YYYY YYYY YMMM MDDD DDTT TTTT TTTT THHH HHMM MMMM SSSS SSmm mmmm mmmm
        # Bits: 5 unused - 12 Year - 4 Month - 5 Day - 11 Timezone - 5 Hour - 6 Minute - 6 Seconds - 10 Milliseconds
        # .....YYYYYYYYYYYYMMMMDDDDDTTTTTTTTTTTHHHHHMMMMMMSSSSSSmmmmmmmmmm
        YearMask = 0b0000011111111111100000000000000000000000000000000000000000000000
        MonthMask = 0b0000000000000000011110000000000000000000000000000000000000000000
        DayMask = 0b0000000000000000000001111100000000000000000000000000000000000000
        # TimezoneMask = 0b0000000000000000000000000011111111111000000000000000000000000000
        HourMask = 0b0000000000000000000000000000000000000111110000000000000000000000
        MinuteMask = 0b0000000000000000000000000000000000000000001111110000000000000000
        SecondsMask = 0b0000000000000000000000000000000000000000000000001111110000000000
        MillisecondsMask = 0b0000000000000000000000000000000000000000000000000000001111111111
        Year = (timeStamp & YearMask) >> 47
        Month = (timeStamp & MonthMask) >> 43
        Day = (timeStamp & DayMask) >> 38
        # Timezone = (timeStamp & TimezoneMask) >> 27
        Hour = (timeStamp & HourMask) >> 22
        Minute = (timeStamp & MinuteMask) >> 16
        Seconds = (timeStamp & SecondsMask) >> 10
        Milliseconds = timeStamp & MillisecondsMask
        logger.debug("Data Timestamp [YYYY-MM-DD HH:MM:SS.mm] = %04u-%02u-%02u %02u:%02u:%02u.%03u" % (
            Year, Month, Day, Hour, Minute, Seconds, Milliseconds))


class CameraParameters:
    """ This class gathers the main camera parameters. """

    def __init__(self, width=176, height=144,
                 cam2worldMatrix=[],
                 fx=146.5, fy=146.5, cx=84.4, cy=71.2,
                 k1=0.326442, k2=0.219623,
                 f2rc=0.0):
        self.width = width
        self.height = height
        self.cam2worldMatrix = [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]
        self.fx = fx
        self.fy = fy
        self.cx = cx
        self.cy = cy
        self.k1 = k1
        self.k2 = k2
        self.f2rc = f2rc


class DepthMap:
    """ This class contains the depthmap data """

    def __init__(self, distance, intensity, confidence, frameNumber, dataQuality, deviceStatus, timestamp):
        self.distance = distance
        self.intensity = intensity
        self.confidence = confidence
        self.frameNumber = frameNumber
        self.dataQuality = dataQuality
        self.deviceStatus = deviceStatus
        self.timestamp = timestamp


class Polar2DData:
    """ This class contains the polar 2D data """

    def __init__(self, angleFirstScanPoint, angularResolution, distance, confidence, timestamp):
        self.distance = distance
        self.angleFirstScanPoint = angleFirstScanPoint
        self.angularResolution = angularResolution
        self.confidence = confidence
        self.timestamp = timestamp


class CartesianData:
    """ This class contains the polar 2D data """

    def __init__(self, numPoints, x, y, z, confidence, timestamp):
        self.numPoints = numPoints
        self.x = x
        self.y = y
        self.z = z
        self.confidence = confidence
        self.timestamp = timestamp
